/*
 * wtmpclean.c -- clear from a wtmp database the unwanted records
 * Copyright (C) 2008 by Davide Madrisan <davide.madrisan@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * indent -i4 -nut src/wtmpclean.c
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif
#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif
#ifdef HAVE_STRING_H
# if !defined STDC_HEADERS && defined HAVE_MEMORY_H
#  include <memory.h>
# endif
# include <string.h>
#endif
#ifdef HAVE_STRINGS_H
# include <strings.h>
#endif

#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <utime.h>

#ifdef HAVE_UTMP_H
# include <utmp.h>
#endif
#ifdef HAVE_UTMPX_H
# include <utmpx.h>
#endif

/* Linux with glibc 2.x:
 * _PATH_UTMPX = /var/run/utmp
 * _PATH_WTMPX = /var/log/wtmp
 */
#ifndef _PATH_WTMPX
# error _PATH_WTMPX is undefined on this platform
#endif

#include "wtmpclean.h"
#include "getopt.h"

void
usage (int status)
{
    static const char *usagemsg[] = {
        PACKAGE " version " PACKAGE_VERSION,
        "Copyright (C) 2008 by Davide Madrisan <davide.madrisan@gmail.com>",
        "",
        "Usage: " PACKAGE " [-l] [-p]"
#ifdef HAVE_UTMPXNAME
            " [-f <wtmpfile>]"
#endif
            " <username>",
#ifdef HAVE_UTMPXNAME
        "  -f, --file       Modify <wtmpfile> instead of " _PATH_WTMPX,
#endif
        "  -l, --list       Show listing of <user> logins",
        "  -p, --preserve   Preserve ownership and timestamps",
        "",
        "Samples:",
#ifdef HAVE_UTMPXNAME
        "  ./" PACKAGE " root",
        "  ./" PACKAGE " -p -f /var/log/wtmp.1 root",
#else
        "  ./" PACKAGE " --preserve root",
#endif
        "",
        "This is free software; see the source for copying conditions.  There is NO",
        "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.",
        (char *) 0
    };
    unsigned int i;

    for (i = 0; usagemsg[i]; i++)
      {
          fprintf (stdout, "%s\n", usagemsg[i]);
      }
    exit (status);
}

int
main (int argc, char **argv)
{
    int opt, opt_index = 0;
    unsigned int lflag = 0, pflag = 0;
    char *wtmpfile = _PATH_WTMPX;
    char *user = NULL;
    unsigned int cleanrec, cleanerr, usrrec;

    static struct option long_options[] = {
#ifdef HAVE_UTMPXNAME
        {"file", required_argument, 0, 'f'},
#endif
        {"list", no_argument, 0, 'l'},
        {"preserve", no_argument, 0, 'p'},
        {"help", no_argument, 0, 'h'},
        {0, 0, 0, 0}
    };
    static const char *options =
#ifdef HAVE_UTMPXNAME
        "f:"
#endif
        "lph";
    opterr = 0;

    while (1)
      {
          opt = getopt_long (argc, argv, options, long_options, &opt_index);
          if (opt == -1)
            {
                break;
            }

          switch (opt)
            {
            case 'f':
                wtmpfile = optarg;
                break;
            case 'p':
                pflag = 1;
                break;
            case 'l':
                lflag = 1;
                break;
            default:
                usage ((opt == 'h') ? 1 : 0);
            }
      }

    if (argc != optind + 1)
      {
          usage (1);
      }

    user = argv[optind];

    if (lflag)
      {
          wtmpchkusr (wtmpfile, user);
          exit (0);
      }

    cleanrec = wtmpedit (wtmpfile, user, pflag, &cleanerr);

    if (cleanerr > 0)
      {
          fprintf (stderr, "Cannot clean up %s\n", wtmpfile);
      }
    else
      {
          printf ("%u block(s) logging user %s removed from %s\n",
                  cleanrec, user, wtmpfile);
      }
}

unsigned int
wtmpedit (const char *wtmpfile, const char *user, unsigned int preserve,
          unsigned int *cleanerr)
{
    struct utmpx *utp;
    unsigned int cleanrec;
    struct stat sb;
    struct utimbuf time;
    uid_t owner;
    gid_t group;

    if (access (wtmpfile, W_OK))
      {
          perror ("Cannot access the file");
          exit (1);
      }

    if (preserve)
      {
          if (stat (wtmpfile, &sb))
            {
                perror ("Cannot get file status");
                exit (1);
            }
          time.actime = sb.st_atime;
          time.modtime = sb.st_mtime;
          owner = sb.st_uid;
          group = sb.st_gid;
      }

#ifdef HAVE_UTMPXNAME
    utmpxname (wtmpfile);
#endif

    setutxent ();

    cleanrec = *cleanerr = 0;
    while (utp = getutxent ())
      {
          if (utp->ut_type == USER_PROCESS &&
              strncmp (utp->ut_user, user, sizeof utp->ut_user) == 0)
            {
                utp->ut_type = DEAD_PROCESS;
                memset (utp->ut_line, 0, sizeof utp->ut_line);
                utp->ut_tv.tv_sec = utp->ut_tv.tv_usec = 0;
                memset (utp->ut_user, 0, sizeof utp->ut_user);
                memset (utp->ut_host, 0, sizeof utp->ut_host);
                memset (utp->ut_addr_v6, 0, sizeof utp->ut_addr_v6);
                /* utp->ut_pid ??? */

                if (pututxline (utp))
                  {
                      cleanrec++;
                  }
                else
                  {
                      cleanerr++;
                  }
            }
      }

    endutxent ();

    if (preserve && chown (wtmpfile, owner, group) < 0)
      {
          fprintf (stderr, "Cannot preserve the ownership of the wtmp file");
      }
    if (preserve && utime (wtmpfile, &time) < 0)
      {
          fprintf (stderr, "Cannot preserve access and modification times");
      }

    return cleanrec;
}

unsigned int
wtmpchkusr (const char *wtmpfile, const char *user)
{
    struct utmpx *utp;
    unsigned int usrrec = 0;
    char *ct;

    if (access (wtmpfile, R_OK))
      {
          perror ("Cannot access the file");
          exit (1);
      }

#ifdef HAVE_UTMPXNAME
    utmpxname (wtmpfile);
#endif

    setutxent ();

    while (utp = getutxent ())
      {
          if (utp->ut_type == USER_PROCESS &&
              strncmp (utp->ut_user, user, sizeof utp->ut_user) == 0)
            {
                ct = ctime (&utp->ut_tv.tv_sec);
                printf ("%3d.  %10.10s %4.4s %5.5s  %s  %s\n", ++usrrec,
                        ct, ct + 20, ct + 11, user, utp->ut_line);
            }
/*
 *        if (strncmp (utp->ut_user, "", sizeof utp->ut_user) == 0)
 *          {
 *              ct = ctime (&utp->ut_tv.tv_sec);
 *              printf ("%10.10s %4.4s %5.5s  logout  %s\n", ct, ct + 20,
 *                      ct + 11, utp->ut_line);
 *          }
 */
      }

    endutxent ();

    return usrrec;
}
